import os
from json import loads
from re import sub

from jaseci.utils.utils import TestCaseHelper
from jaseci.jsorc.jsorc import JsOrc
from jaseci.utils.test_core import skip_without_redis

from django.contrib.auth import get_user_model
from django.urls import reverse
from django.test import TestCase
from rest_framework.test import APIClient
from django_celery_beat.models import PeriodicTask, CrontabSchedule
from jaseci_serv.base.models import JaseciObject


class TaskServiceTest(TestCaseHelper, TestCase):
    def assert_obj(self, val=None):
        # check on db if committed
        self.assertEqual(
            val,
            loads(JaseciObject.objects.get(jid=self.node_a).jsci_obj)["context"]["val"],
        )

    def setUp(self):
        super().setUp()
        # First user is always super,
        self.admin = get_user_model().objects.create_superuser(
            "admin@jaseci.com", "password"
        )
        self.admin_client = APIClient()
        self.admin_client.force_authenticate(self.admin)
        self.admin_master = self.admin.get_master()

        self.user = get_user_model().objects.create_user("user@jaseci.com", "password")
        self.user_client = APIClient()
        self.user_client.force_authenticate(self.user)
        self.user_master = self.user.get_master()

        zsb_file = open(os.path.dirname(__file__) + "/testing.jac").read()
        payload = {"op": "sentinel_register", "name": "public", "code": zsb_file}
        self.admin_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )

        payload = {"op": "walker_run", "name": "init"}
        self.node_a = self.admin_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        ).data["report"][0]["jid"]

        self.assert_obj()

        payload = {"op": "global_sentinel_set"}
        self.admin_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )

        payload = {"op": "sentinel_active_global"}
        self.admin_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )

        # Simulate celery.backend_cleanup
        PeriodicTask.objects.create(
            name="celery.backend_cleanup",
            task="celery.backend_cleanup",
            crontab=CrontabSchedule.objects.get_or_create(
                minute="0", hour="4", day_of_week="*"
            )[0],
        )

    @skip_without_redis
    def test_add_scheduled_walker(self):
        res = self.admin_client.post(
            reverse(f'jac_api:{"add_scheduled_walker"}'),
            {
                "name": "scheduled_walker",
                "schedule": {"type": "interval", "every": 1, "period": "minutes"},
                "body": {"wlk": "scheduled_walker"},
            },
            format="json",
        ).data

        self.assertEqual("Scheduled Walker created successfully!", res)

        periodic_task = PeriodicTask.objects.get(name="scheduled_walker")
        self.assertIsNotNone(periodic_task)
        self.assertEqual("scheduled_walker", periodic_task.name)

        # running the actual periodic task
        args = {
            "args": loads(periodic_task.args),
            "kwargs": loads(periodic_task.kwargs),
            "periodic_task_name": periodic_task.name,
        }

        if periodic_task.queue:
            args["queue"] = periodic_task.queue

        task_svc = JsOrc.svc("task")
        task_result = (
            task_svc.app.tasks.get(periodic_task.task).apply_async(**args).result
        )

        self.assertTrue(task_result["success"])
        self.assertEqual([True], task_result["report"])
        self.assert_obj(True)

        res = self.admin_client.post(
            reverse(f'jac_api:{"get_scheduled_queues"}'),
            {},
            format="json",
        ).data

        self.assertEqual(
            res,
            [
                {
                    "id": 2,
                    "name": "scheduled_walker",
                    "schedule": {
                        "type": "interval",
                        "conf": {"every": 1, "period": "minutes"},
                    },
                    "kwargs": {"wlk": "scheduled_walker", "mst": self.admin_master.jid},
                }
            ],
        )

        res = self.admin_client.post(
            reverse(f'jac_api:{"delete_scheduled_queue"}'),
            {"scheduled_queue_id": periodic_task.id},
            format="json",
        ).data

        self.assertEqual("Successfully deleted!", res)

    @skip_without_redis
    def test_add_scheduled_sequence(self):
        res = self.admin_client.post(
            reverse(f'jac_api:{"add_scheduled_sequence"}'),
            {
                "name": "scheduled_sequence",
                "schedule": {"type": "interval", "every": 1, "period": "minutes"},
                "body": {
                    "persistence": {
                        "additional_field": "can_be_call_via_#.additional_field",
                        "master": f"{self.admin_master.jid}",
                    },
                    "requests": [
                        {
                            "method": "JAC",
                            "api": "master_allusers",
                            "master": "{{#.master}}",
                            "body": {"limit": 0, "offset": 0},
                            "ignore_error": True,
                            "__def_loop__": {
                                "by": "$.data",
                                "filter": [
                                    {
                                        "or": [
                                            {
                                                "by": "$.user",
                                                "condition": {
                                                    "eq": "user@jaseci.com",
                                                    "ne": None,
                                                    "gt": None,
                                                    "gte": None,
                                                    "lt": None,
                                                    "lte": None,
                                                    "regex": None,
                                                },
                                            },
                                            {
                                                "and": [
                                                    {
                                                        "by": "$.user",
                                                        "condition": {
                                                            "eq": "user@jaseci.com"
                                                        },
                                                    }
                                                ]
                                            },
                                        ]
                                    }
                                ],
                                "requests": [
                                    {
                                        "method": "JAC",
                                        "api": "object_get",
                                        "master": "{{#.master}}",
                                        "body": {
                                            "obj": "{{$.jid}}",
                                            "depth": 0,
                                            "detailed": True,
                                        },
                                        "header": {
                                            "Authorization": "token {{#.login.token}}"
                                        },
                                        "ignore_error": True,
                                        "save_to": "object_get_{{!}}",
                                    },
                                    {
                                        "method": "JAC",
                                        "api": "walker_run",
                                        "master": "{{#.master}}",
                                        "body": {
                                            "name": "scheduled_sequence",
                                            "ctx": {},
                                            "nd": "{{$.active_gph_id}}",
                                            "snt": "active:sentinel",
                                        },
                                        "header": {
                                            "Authorization": "token {{#.login.token}}"
                                        },
                                        "save_to": "response_{{@.jid}}",
                                        "save_req_to": "req_{{@.jid}}",
                                    },
                                ],
                            },
                        },
                        {
                            "method": "GET",
                            "api": "https://jsonplaceholder.typicode.com/todos/100",
                            "save_to": "testing_nested",
                            "save_req_to": "req_testing_nested",
                        },
                        {
                            "method": "JAC",
                            "api": "master_allusers",
                            "master": "{{#.master}}",
                            "body": {"limit": 0, "offset": 0},
                            "save_to": "all_users",
                        },
                    ],
                },
            },
            format="json",
        ).data

        self.assertEqual("Scheduled Sequence created successfully!", res)

        periodic_task = PeriodicTask.objects.get(name="scheduled_sequence")
        self.assertIsNotNone(periodic_task)
        self.assertEqual("scheduled_sequence", periodic_task.name)

        # running the actual periodic task
        args = {
            "args": loads(periodic_task.args),
            "kwargs": loads(periodic_task.kwargs),
            "periodic_task_name": periodic_task.name,
        }

        if periodic_task.queue:
            args["queue"] = periodic_task.queue

        task_svc = JsOrc.svc("task")
        task_result = (
            task_svc.app.tasks.get(periodic_task.task).apply_async(**args).result
        )

        self.assertEqual(
            "can_be_call_via_#.additional_field", task_result["additional_field"]
        )
        self.assertEqual(self.admin_master.jid, task_result["master"])
        self.assertEqual("user@jaseci.com", task_result["object_get_0"]["name"])
        self.assertEqual("Jaseci Master", task_result["object_get_0"]["kind"])
        self.assertEqual(self.user_master.jid, task_result["object_get_0"]["jid"])

        key = sub(r"[\:\-]", "_", self.user_master.jid)

        self.assertTrue((f"req_{key}") in task_result)
        self.assertFalse(task_result[f"response_{key}"]["report"][0])
        self.assertTrue("req_testing_nested" in task_result)
        self.assertTrue("testing_nested" in task_result)
        self.assertEqual("user@jaseci.com", task_result["all_users"]["data"][0]["user"])
        self.assertEqual(
            "admin@jaseci.com", task_result["all_users"]["data"][1]["user"]
        )
        self.assert_obj(False)

    @skip_without_redis
    def test_add_scheduled_sequence_via_user(self):
        res = self.user_client.post(
            reverse(f'jac_api:{"add_scheduled_sequence"}'),
            {
                "name": "scheduled_sequence",
                "schedule": {"type": "interval", "every": 1, "period": "minutes"},
                "body": {
                    "persistence": {
                        "additional_field": "can_be_call_via_#.additional_field",
                        "master": f"{self.admin_master.jid}",
                    },
                    "requests": [],
                },
            },
            format="json",
        ).data

        self.assertEqual(
            "You do not have permission to perform this action.", str(res["detail"])
        )



import os
from django.test import TestCase, Client, override_settings
from django.urls import reverse
from jaseci_serv.settings import BASE_DIR


class StudioRoutesTestWithTemplates(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.override_settings = override_settings(
            TEMPLATES=[
                {
                    "BACKEND": "django.template.backends.django.DjangoTemplates",
                    "APP_DIRS": True,
                    "DIRS": [
                        os.path.join(
                            BASE_DIR, "jaseci_serv/studio/tests/mocks_templates/"
                        ),
                    ],
                }
            ]
        )
        cls.override_settings.__enter__()

    @classmethod
    def tearDownClass(cls):
        print("Tearing down class")
        cls.override_settings.__exit__(None, None, None)
        super().tearDownClass()

    def setUp(self):
        self.client = Client()

    def test_studio_index(self):
        response = self.client.get(reverse("studio_index"))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "studio/index.html")

    def test_studio_dashboard(self):
        response = self.client.get(reverse("studio_dashboard"))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "studio/dashboard.html")

    def test_studio_graph_viewer(self):
        response = self.client.get(reverse("studio_graph_viewer"))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "studio/graph-viewer.html")

    def test_studio_logs(self):
        response = self.client.get(reverse("studio_logs_viewer"))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "studio/logs.html")

    def test_studio_architype(self):
        response = self.client.get(reverse("studio_architype"))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "studio/architype.html")

    def test_studio_actions(self):
        response = self.client.get(reverse("studio_actions"))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "studio/actions.html")


class StudioRoutesTestWithoutTemplates(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.override_settings = override_settings(
            TEMPLATES=[
                {
                    "BACKEND": "django.template.backends.django.DjangoTemplates",
                    "APP_DIRS": True,
                    "DIRS": [
                        os.path.join(
                            BASE_DIR,
                            "jaseci_serv/studio/tests/missing_mocks_templates/",
                        ),
                    ],
                }
            ]
        )
        cls.override_settings.__enter__()

    @classmethod
    def tearDownClass(cls):
        cls.override_settings.__exit__(None, None, None)
        super().tearDownClass()

    def setUp(self):
        self.client = Client()

    def test_studio_index(self):
        response = self.client.get(reverse("studio_index"))
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.content, b"Jaseci Studio is not installed.")

    def test_studio_graph_viewer(self):
        response = self.client.get(reverse("studio_graph_viewer"))
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.content, b"Jaseci Studio is not installed.")

    def test_studio_logs(self):
        response = self.client.get(reverse("studio_logs_viewer"))
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.content, b"Jaseci Studio is not installed.")

    def test_studio_architype(self):
        response = self.client.get(reverse("studio_architype"))
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.content, b"Jaseci Studio is not installed.")

    def test_studio_actions(self):
        response = self.client.get(reverse("studio_actions"))
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.content, b"Jaseci Studio is not installed.")



from json import dumps, loads
from jaseci.jsorc.jsorc import JsOrc
from jaseci.extens.svc.task_svc import TaskService as Ts
from django.db.models import Q
from django_celery_beat.models import PeriodicTask, CrontabSchedule, IntervalSchedule

#################################################
#                 TASK APP ORM                 #
#################################################


@JsOrc.service(name="task", config="TASK_CONFIG", priority=2)
class TaskService(Ts):
    ###################################################
    #                SCHEDULED QUEUING                #
    ###################################################

    def add_scheduled_queue(
        self, queue_type: object, name: str, schedule: dict, body: dict
    ):
        periodic_task = {
            "name": name,
            "task": self.get_task_name(queue_type),
            "kwargs": dumps(body),
            "one_off": schedule.get("one_off", False),
        }

        if schedule.get("type") == "interval":
            interval = {"every": 1, "period": IntervalSchedule.MINUTES}
            interval.update(schedule.get("conf", {}))
            periodic_task["interval"] = IntervalSchedule.objects.get_or_create(
                **interval
            )[0]
        else:
            crontab = {
                "minute": "*",
                "hour": "*",
                "day_of_week": "*",
                "day_of_month": "*",
                "month_of_year": "*",
            }
            crontab.update(schedule.get("conf", {}))
            periodic_task["crontab"] = CrontabSchedule.objects.get_or_create(**crontab)[
                0
            ]

        return PeriodicTask.objects.create(**periodic_task)

    def get_scheduled_queues(
        self,
        limit: int = 10,
        offset: int = 0,
        asc: bool = True,
        name: str = None,
        master=False,
    ):
        from jaseci.prim.super_master import SuperMaster

        is_admin = isinstance(master, SuperMaster)

        filter = [Q(crontab__isnull=False) | Q(interval__isnull=False), ~Q(id=1)]
        if not is_admin:
            filter.append(Q(task=self.get_task_name(self.scheduled_walker)))
        if name:
            filter.append(Q(name__icontains=name))

        pts = []
        for pt in PeriodicTask.objects.filter(*filter).order_by(
            ("" if asc else "-") + "name"
        ):
            kwargs = loads(pt.kwargs)
            mst = kwargs.get("mst")
            try:
                if is_admin or (mst != None and mst in master.jid):
                    if pt.interval:
                        schedule = {
                            "type": "interval",
                            "conf": {
                                "every": pt.interval.every,
                                "period": pt.interval.period,
                            },
                        }
                    else:
                        schedule = {
                            "type": "crontab",
                            "conf": {
                                "minute": pt.crontab.minute,
                                "hour": pt.crontab.hour,
                                "day_of_week": pt.crontab.day_of_week,
                                "day_of_month": pt.crontab.day_of_month,
                                "month_of_year": pt.crontab.month_of_year,
                            },
                        }

                    pts.append(
                        {
                            "id": pt.id,
                            "name": pt.name,
                            "schedule": schedule,
                            "kwargs": loads(pt.kwargs),
                        }
                    )
            except Exception:
                pass

        return pts[offset : offset + limit if limit else PeriodicTask.objects.count()]

    def delete_scheduled_queue(self, scheduled_queue_id: int, master):
        from jaseci.prim.super_master import SuperMaster

        is_admin = isinstance(master, SuperMaster)

        try:
            pt = PeriodicTask.objects.get(id=scheduled_queue_id)
            if pt:
                kwargs = loads(pt.kwargs)
                mst = kwargs.get("mst")
                if is_admin or (mst != None and mst in master.jid):
                    pt.delete()
                    return True
        except PeriodicTask.DoesNotExist:
            pass

        return False



import os
import pytest
from django.contrib.auth import get_user_model
from django.urls import reverse

from rest_framework.test import APIClient

from jaseci.utils.utils import TestCaseHelper
from django.test import TestCase

from jaseci.jsorc.jsorc import JsOrc
from kubernetes import config as kubernetes_config, client
from unittest.mock import patch, Mock
from jaseci.jsorc.jsorc_settings import JsOrcSettings
from jaseci.extens.svc.elastic_svc import LOG_QUEUES, format_elastic_record


class LoggingTests(TestCaseHelper, TestCase):
    """Test logging with jac API request"""

    @patch.object(client.ApiClient, "call_api")
    @patch.object(kubernetes_config, "load_incluster_config")
    def setUp(self, mock_config, mock_call_api):
        super().setUp()
        self.client = APIClient()
        get_user_model().objects.create_user(
            "throwawayJSCITfdfdEST_test@jaseci.com", "password"
        )
        self.user = get_user_model().objects.create_user(
            "JSCITfdfdEST_test@jaseci.com", "password"
        )
        self.auth_client = APIClient()
        self.auth_client.force_authenticate(self.user)
        self.suser = get_user_model().objects.create_superuser(
            "JSCITfdfdEST_test2@jaseci.com", "password", name="Administrator"
        )
        self.sauth_client = APIClient()
        self.sauth_client.force_authenticate(self.suser)

        JsOrcSettings.KUBE_CONFIG["enabled"] = True
        JsOrcSettings.ELASTIC_CONFIG["enabled"] = True
        JsOrcSettings.ELASTIC_CONFIG["under_test"] = True
        self.kube = JsOrc.svc("kube")
        with patch.object(self.kube, "read") as mock_read, patch.object(
            self.kube, "patch"
        ) as mock_patch, patch.object(self.kube, "create") as mock_create, patch(
            "jaseci.extens.svc.elastic_svc.Elastic._get"
        ) as mocked_get:
            mocked_get.return_value = {"timed_out": False}
            mock_read.return_value = Mock()
            mock_patch.return_value = Mock()
            mock_create.return_value = Mock()
            JsOrc.svc_reset("elastic")

    def tearDown(self):
        super().tearDown()
        JsOrcSettings.KUBE_CONFIG["enabled"] = False
        JsOrcSettings.ELASTIC_CONFIG["enabled"] = False
        JsOrcSettings.ELASTIC_CONFIG["under_test"] = False
        JsOrc.svc_reset("elastic")
        JsOrc.svc_reset("kube")
        self.logger_off()

    def test_elastic_logging_objects(self):
        """Test sentinel register and walker run is logged properly"""
        self.logger_on()
        jac_file = open(os.path.dirname(__file__) + "/general.jac").read()
        payload = {"op": "sentinel_register", "name": "general", "code": jac_file}
        self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )

        # There should be two entries in the log queue
        log_queue = LOG_QUEUES["core"]
        self.assertEqual(log_queue.qsize(), 2)

        # Validate the first entry
        record = log_queue.get_nowait()
        elastic_record = format_elastic_record(record)

        self.assertEqual(
            set(elastic_record.keys()),
            set(
                [
                    "@timestamp",
                    "message",
                    "level",
                    "api_name",
                    "caller_name",
                    "caller_jid",
                    "request_user_agent",
                    "request_payload",
                ]
            ),
        )

        # Validate the second entry
        record = log_queue.get_nowait()
        elastic_record = format_elastic_record(record)

        self.assertEqual(
            set(elastic_record.keys()),
            set(
                [
                    "@timestamp",
                    "message",
                    "level",
                    "api_name",
                    "request_latency",
                    "objects_touched",
                    "redis_touches",
                    "db_touches",
                    "objects_touched_size",
                    "objects_saved",
                    "caller_name",
                    "caller_jid",
                    "api_response",
                ]
            ),
        )

        # Validate walker run
        # Walker run will have walker_name and success in the log record
        payload = {"op": "walker_run", "name": "here_fix_create"}
        self.auth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )

        # There should be two entries in the log queue
        log_queue = LOG_QUEUES["core"]
        self.assertEqual(log_queue.qsize(), 2)

        # Validate the first entry
        record = log_queue.get_nowait()
        elastic_record = format_elastic_record(record)

        self.assertEqual(
            set(elastic_record.keys()),
            set(
                [
                    "@timestamp",
                    "message",
                    "level",
                    "api_name",
                    "caller_name",
                    "caller_jid",
                    "request_user_agent",
                    "request_payload",
                    "walker_name",
                ]
            ),
        )

        # Validate the second entry
        record = log_queue.get_nowait()
        elastic_record = format_elastic_record(record)

        self.assertEqual(
            set(elastic_record.keys()),
            set(
                [
                    "@timestamp",
                    "message",
                    "level",
                    "api_name",
                    "request_latency",
                    "objects_touched",
                    "redis_touches",
                    "db_touches",
                    "objects_touched_size",
                    "objects_saved",
                    "caller_name",
                    "caller_jid",
                    "api_response",
                    "walker_name",
                    "success",
                ]
            ),
        )

        self.logger_off()



from django.urls import path
from jaseci_serv.studio.views import *


urlpatterns = [
    path(
        "",
        studio_index,
        name="studio_index",
    ),
    path(
        "dashboard/",
        studio_dashboard,
        name="studio_dashboard",
    ),
    path(
        "graph-viewer/",
        studio_graph_viewer,
        name="studio_graph_viewer",
    ),
    path(
        "logs/",
        studio_logs_viewer,
        name="studio_logs_viewer",
    ),
    path(
        "architype/",
        studio_architype,
        name="studio_architype",
    ),
    path(
        "actions/",
        studio_actions,
        name="studio_actions",
    ),
]



from django import forms
from django.contrib import admin

from .models import SocialApp, LegacySocialApp


class SocialAppForm(forms.ModelForm):
    class Meta:
        model = SocialApp
        exclude = []
        widgets = {
            "client_id": forms.TextInput(attrs={"size": "100"}),
            "key": forms.TextInput(attrs={"size": "100"}),
            "secret": forms.TextInput(attrs={"size": "100"}),
        }


class SocialAppAdmin(admin.ModelAdmin):
    form = SocialAppForm
    list_display = ("id", "name", "provider")


admin.site.unregister(LegacySocialApp)
admin.site.register(SocialApp, SocialAppAdmin)



from jaseci.extens.api.interface import Interface
from jaseci.extens.api.jsorc_api import JsOrcApi as CoreJsOrcApi
from jaseci_serv.base.jsorc_loadtest import JsorcLoadTest


class JsOrcApi(CoreJsOrcApi):
    @Interface.admin_api()
    def jsorc_loadtest(
        self,
        test: str,
        experiment: str = "",
        mem: int = 0,
        policy: str = "all_local",
        experiment_duration: int = 180,
        eval_phase: int = 10,
        perf_phase: int = 100,
    ):
        """
        A jsorc loadtest
        """
        tester = JsorcLoadTest(test)
        return tester.run_test(
            experiment, mem, policy, experiment_duration, eval_phase, perf_phase
        )



from rest_framework import generics, permissions
from rest_framework.views import APIView
from rest_framework import renderers, status
from knox.views import LoginView as KnoxLoginView
from django.contrib.auth import login, get_user_model
from django.contrib.auth.signals import user_logged_out
from knox.auth import TokenAuthentication

from jaseci_serv.user_api.serializers import UserSerializer
from jaseci_serv.user_api.serializers import SuperUserSerializer
from jaseci_serv.user_api.serializers import AuthTokenSerializer
from jaseci_serv.base.models import lookup_global_config
from datetime import timedelta

from rest_framework.response import Response
import base64

from jaseci.jsorc.jsorc import JsOrc
from jaseci.utils.utils import logger, ColCodes as Cc


class CreateUserView(generics.CreateAPIView):
    """Create a new user in the system"""

    serializer_class = UserSerializer


class CreateSuperUserView(generics.CreateAPIView):
    """Create a new user in the system"""

    serializer_class = SuperUserSerializer


class ActivateUserView(APIView):
    """Create a new user in the system"""

    # TODO: Add Test case
    renderer_classes = [renderers.JSONRenderer]

    def get(self, request, code):
        """Perform create override to send out activation Email"""
        email = base64.b64decode(code).decode()
        user = get_user_model().objects.get(email=email)
        if not user:
            return Response("Invalid activation code/link!")
        user.is_activated = True
        user.save()
        return Response("User successfully activated!")


class CreateTokenView(KnoxLoginView):
    """Create a new auth token for user"""

    permission_classes = (permissions.AllowAny,)

    def get_token_ttl(self):
        ttl = lookup_global_config("LOGIN_TOKEN_TTL_HOURS", None)
        if not ttl:
            return super(CreateTokenView, self).get_token_ttl()
        else:
            return timedelta(hours=int(ttl))

    def post(self, request, format=None):
        serializer = AuthTokenSerializer(
            data=request.data, context={"request": request}
        )
        serializer.is_valid(raise_exception=True)
        user = serializer.validated_data["user"]
        user_agent = request.META.get("HTTP_USER_AGENT", "")
        log_str = f"Login request from {Cc.TG}{user}{Cc.EC} via {user_agent}"
        log_dict = {
            "api_name": "login",
            "caller_name": str(user),
            "request_user_agent": user_agent,
        }
        log_dict["extra_fields"] = list(log_dict.keys())
        logger.info(log_str, extra=log_dict)

        login(request, user)
        res = super(CreateTokenView, self).post(request, format=None)

        log_str = f"Login success for {Cc.TG}{user}{Cc.EC} via {user_agent}"
        log_dict = {
            "api_name": "login",
            "caller_name": str(user),
            "request_user_agent": user_agent,
        }
        log_dict["extra_fields"] = list(log_dict.keys())
        logger.info(log_str, extra=log_dict)
        return res


class ManageUserView(generics.RetrieveUpdateAPIView):
    """Manage the authenticated user"""

    # TODO: re-activate when user changes email address or disable email change
    serializer_class = UserSerializer
    authentication_classes = (TokenAuthentication,)
    permission_classes = (permissions.IsAuthenticated,)

    def get_object(self):
        """Retrieve and return authenticated user"""
        return self.request.user

    def process_update(self, request, *args, **kwargs):
        user = request.user
        current_user = {
            "id": user.id,
            "email": user.email,
            "name": user.name,
            "is_activated": user.is_activated,
            "is_superuser": user.is_superuser,
        }

        response = self.update(request, *args, **kwargs)

        elastic = JsOrc.svc("elastic")
        if elastic.is_running():
            activity = elastic.app.generate_from_request(request)

            user = response.data

            activity["misc"] = {
                "old": current_user,
                "new": {
                    "id": user["id"],
                    "email": user["email"],
                    "name": user["name"],
                    "is_activated": user["is_activated"],
                    "is_superuser": user["is_superuser"],
                },
            }
            elastic.app.doc_activity(activity)

        return response

    def put(self, request, *args, **kwargs):
        return self.process_update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        kwargs["partial"] = True
        return self.process_update(request, *args, **kwargs)


class LogoutAllUsersView(APIView):
    """
    Log out of all user sessions across all users
    I.E. deletes all auth tokens for all users
    """

    authentication_classes = (TokenAuthentication,)
    permission_classes = (permissions.IsAuthenticated, permissions.IsAdminUser)

    def post(self, request, format=None):
        users = get_user_model().objects.all()
        for u in users:
            u.auth_token_set.all().delete()
            user_logged_out.send(sender=u.__class__, request=request, user=u)
        return Response(None, status=status.HTTP_204_NO_CONTENT)


class IsSuperUser(permissions.BasePermission):
    """
    Allows access only to super users.
    """

    def has_permission(self, request, view):
        return bool(request.user and request.user.is_superuser)


class UpdateUserView(APIView):
    """
    update specific user's is_activated and is_superuser
    """

    serializer_class = UserSerializer
    authentication_classes = (TokenAuthentication,)
    permission_classes = (permissions.IsAuthenticated, IsSuperUser)

    def update_field(self, user, data: dict, field_map: dict):
        has_changes = False
        for field, _type in field_map.items():
            if field in data and type(data[field]) is _type:
                if getattr(user, field, _type()) != data[field]:
                    setattr(user, field, data[field])
                    has_changes = True

        return has_changes

    def post(self, request, id):
        user = get_user_model().objects.get(id=id)

        current_user = {
            "id": user.id,
            "email": user.email,
            "name": user.name,
            "is_activated": user.is_activated,
            "is_superuser": user.is_superuser,
        }

        if user is None or not request.data or type(request.data) is not dict:
            return Response(status=status.HTTP_400_BAD_REQUEST)

        if self.update_field(
            user, request.data, {"is_activated": bool, "is_superuser": bool}
        ):
            user.save()

            elastic = JsOrc.svc("elastic")
            if elastic.is_running():
                activity = elastic.app.generate_from_request(request)

                activity["misc"] = {
                    "old": current_user,
                    "new": {
                        "id": user.id,
                        "email": user.email,
                        "name": user.name,
                        "is_activated": user.is_activated,
                        "is_superuser": user.is_superuser,
                    },
                }
                elastic.app.doc_activity(activity)

            return Response("Update Success!")
        else:
            return Response("No changes found!")



import base64

from django.contrib.auth import get_user_model
from django.urls import reverse
from jaseci.utils.test_core import skip_without_redis

from rest_framework.test import APIClient
from rest_framework import status

from jaseci.utils.utils import TestCaseHelper
from django.test import TestCase

import uuid
import os


class JacGeneralTests(TestCaseHelper, TestCase):
    """Test the publicly available node API"""

    def setUp(self):
        super().setUp()
        self.client = APIClient()
        self.user = get_user_model().objects.create_user(
            "JSCITfdfdEST_test@jaseci.com", "password"
        )
        self.client.force_authenticate(self.user)

    def tearDown(self):
        super().tearDown()

    def quick_call(self, bywho, ops):
        return bywho.post(reverse(f'jac_api:{ops["op"]}'), ops, format="json")

    def test_here_update_fix(self):
        testfile = open(os.path.dirname(__file__) + "/general.jac").read()
        self.quick_call(
            self.client, {"op": "sentinel_register", "name": "test", "code": testfile}
        )

        self.user._h.clear_cache()
        ret = self.quick_call(
            self.client, {"op": "walker_run", "name": "here_fix_create"}
        )
        self.user._h.clear_cache()
        ret = self.quick_call(
            self.client,
            {
                "op": "walker_run",
                "name": "here_fix_update",
                "nd": ret.data["report"][0],
            },
        )
        self.user._h.clear_cache()
        ret = self.quick_call(
            self.client, {"op": "walker_run", "name": "here_fix_check"}
        )
        self.user._h.clear_cache()
        self.assertEqual(ret.data["report"], [{"value": "Test"}])

    def test_object_bug(self):
        testfile = open(os.path.dirname(__file__) + "/object_bug.jac").read()
        self.quick_call(
            self.client, {"op": "sentinel_register", "name": "test", "code": testfile}
        )
        self.user._h.clear_cache()
        ret = self.quick_call(self.client, {"op": "walker_run", "name": "test_walk"})
        self.assertEqual(ret.data["success"], True)
        self.user._h.clear_cache()
        ret = self.quick_call(self.client, {"op": "walker_run", "name": "test_walk"})
        self.assertEqual(ret.data["success"], True)
        self.user._h.clear_cache()
        ret = self.quick_call(self.client, {"op": "walker_run", "name": "test_walk"})
        self.assertEqual(ret.data["success"], True)



import os
from django.apps import AppConfig

from jaseci.jsorc.jsorc import JsOrc


class CoreApiConfig(AppConfig):
    name = "jaseci_serv.svc"

    def ready(self):
        from .task_svc import TaskService  # noqa

        if os.environ.get("RUN_MAIN"):
            JsOrc.run()



import random
import string
from allauth.socialaccount.providers.google.views import GoogleOAuth2Adapter
from allauth.socialaccount.providers.facebook.views import FacebookOAuth2Adapter
from allauth.socialaccount.providers.github.views import GitHubOAuth2Adapter
from allauth.socialaccount.providers.microsoft.views import MicrosoftGraphOAuth2Adapter
from allauth.socialaccount.providers.okta.views import OktaOAuth2Adapter
from allauth.socialaccount.providers.oauth2.client import OAuth2Client
from allauth.socialaccount.providers.apple.views import (
    AppleOAuth2Adapter as LegacyAppleOAuth2Adapter,
    AppleOAuth2Client,
)
from allauth.socialaccount.adapter import DefaultSocialAccountAdapter, get_adapter
from rest_framework.generics import GenericAPIView
from rest_framework.response import Response
from jaseci_serv.jsx_oauth.utils import GetExampleUrlSerializer
from django.views import View
from django.shortcuts import render
from django.utils.translation import gettext_lazy as _
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from rest_framework import status
from rest_framework.generics import GenericAPIView
from rest_framework.response import Response
from jaseci_serv.jsx_oauth.models import (
    SocialLoginProvider,
    PROVIDERS_MAPPING,
    SocialApp,
)
from jaseci_serv.jsx_oauth.utils import JSXSocialLoginView

#################################################
#     OVERRIDING DEFAULTS FOR APPLE SIGN IN     #
#################################################

main_get_app = DefaultSocialAccountAdapter.get_app


def custom_get_app(self, request, provider, config=None):
    if hasattr(request, "app") and request.app:
        return request.app
    return main_get_app(self, request, provider, config)


DefaultSocialAccountAdapter.get_app = custom_get_app


class AppleOAuth2Adapter(LegacyAppleOAuth2Adapter):
    def get_client_id(self, provider):
        app = get_adapter().get_app(self.request, provider)
        return [aud.strip() for aud in app.client_id.split(",")]


# --------------------------------------------- #


def generate_state():
    return "".join(
        random.choices(
            string.ascii_uppercase + string.ascii_lowercase + string.digits, k=64
        )
    )


class ExampleUrlView(GenericAPIView):
    serializer_class = GetExampleUrlSerializer

    def post(self, request):
        serializer_data = self.serializer_class(data=request.data)
        serializer_data.is_valid(raise_exception=True)
        provider = serializer_data.validated_data.get("provider")
        provider_map = PROVIDERS_MAPPING[provider]
        callback_url = f'{request.build_absolute_uri("/")[:-1]}{provider_map["DEFAULT_REDIRECT_URI"]}'
        social_app = SocialApp.objects.filter(provider=provider.lower()).first()
        if social_app:
            return Response(
                {
                    "url": provider_map["LOGIN_URL"].format(
                        callback_url=callback_url,
                        client_id=social_app.client_id,
                        state=generate_state(),
                    )
                }
            )
        else:
            return Response(
                {
                    "error": f"Social login applications for the {provider} not configured properly"
                },
                status=status.HTTP_400_BAD_REQUEST,
            )


class GoogleLogin(JSXSocialLoginView):
    adapter_class = GoogleOAuth2Adapter
    provider = SocialLoginProvider.GOOGLE
    client_class = OAuth2Client


class FaceBookLogin(JSXSocialLoginView):
    adapter_class = FacebookOAuth2Adapter
    provider = SocialLoginProvider.FACEBOOK
    client_class = OAuth2Client


class GitHubLogin(JSXSocialLoginView):
    adapter_class = GitHubOAuth2Adapter
    provider = SocialLoginProvider.GITHUB
    client_class = OAuth2Client


class OktaLogin(JSXSocialLoginView):
    adapter_class = OktaOAuth2Adapter
    provider = SocialLoginProvider.OKTA
    client_class = OAuth2Client


class MicrosoftLogin(JSXSocialLoginView):
    adapter_class = MicrosoftGraphOAuth2Adapter
    provider = SocialLoginProvider.MICROSOFT
    client_class = OAuth2Client


class OpenIdLogin(JSXSocialLoginView):
    adapter_class = GitHubOAuth2Adapter  # TODO: create custom adaptor for OpenID
    provider = SocialLoginProvider.OPENID
    client_class = OAuth2Client


class AppleLogin(JSXSocialLoginView):
    adapter_class = AppleOAuth2Adapter
    provider = SocialLoginProvider.APPLE
    client_class = AppleOAuth2Client


@method_decorator(csrf_exempt, name="dispatch")
class ExampleLiveView(View):
    template_name = "examples/social_auth.html"

    def get(self, request, provider):
        code = request.GET.get("code")
        social_app = SocialApp.objects.filter(provider=provider).first()
        return render(
            request,
            self.template_name,
            {
                "provider": provider,
                "code": code,
                "client_id": social_app.client_id,
                "state": generate_state(),
            },
        )

    def post(self, request, provider):
        data = request.POST.dict()
        return render(
            request,
            self.template_name,
            {
                "provider": provider,
                "id_token": data["id_token"],
                "state": data["state"],
            },
        )



from jaseci.jsorc.jsorc import JsOrc
from jaseci.utils.utils import logger, ColCodes as Cc
from jaseci_serv.jsx_oauth.models import PROVIDERS_MAPPING, SocialLoginProvider
from allauth.socialaccount.providers.oauth2.client import OAuth2Client
from dj_rest_auth.registration.views import SocialLoginView
from dj_rest_auth.registration.serializers import SocialLoginSerializer
from jaseci_serv.base.models import lookup_global_config
from django.core.exceptions import ObjectDoesNotExist
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from knox.settings import knox_settings
from knox.models import AuthToken
from django.contrib.auth import authenticate, get_user_model
from datetime import timedelta
from rest_framework.response import Response
from rest_framework.exceptions import APIException
from rest_framework import serializers
from allauth.socialaccount.helpers import complete_social_login
from requests.exceptions import HTTPError
from enum import Enum
from time import time
from json import dumps

from .models import SocialApp

OBJECT_LOG_LIMIT = 5 * 1024 * 1024


class LoginType(Enum):
    REGISTRATION = "New account"
    EXISTING = "Existing account"
    ACTIVATION = "Existing account but just got activated"
    BINDING = "Existing account but just got binded with different login type"


class AppIdFieldNotExists(APIException):
    status_code = 400
    default_detail = "app_id is required!"
    default_code = "Request failed!"


class GetExampleUrlSerializer(serializers.Serializer):
    provider = serializers.ChoiceField(choices=SocialLoginProvider.choices)


class JSXSocialLoginSerializer(SocialLoginSerializer):
    app_id = serializers.CharField(required=False, allow_blank=True)

    # deprecated
    internal_client_id = serializers.CharField(required=False, allow_blank=True)

    def validate(self, attrs):
        view = self.context.get("view")
        request = self._get_request()

        if not view:
            raise serializers.ValidationError(
                _("View is not defined, pass it as a context variable"),
            )

        adapter_class = getattr(view, "adapter_class", None)
        if not adapter_class:
            raise serializers.ValidationError(_("Define adapter_class in view"))

        adapter = adapter_class(request)

        # internal_client_id is deprecated
        app_id = attrs.get("app_id") or attrs.get("internal_client_id")

        if app_id:
            try:
                app = SocialApp.objects.get(id=app_id).legacy()
            except SocialApp.DoesNotExist:
                raise serializers.ValidationError(
                    _(f"app_id is not yet associated on any Social App!"),
                )
        else:
            social_apps = SocialApp.objects.filter(provider=adapter.provider_id)
            social_apps_count = len(social_apps)

            if social_apps_count > 1:
                raise serializers.ValidationError(
                    _(
                        f"You have multiple {adapter.provider_id} Social App. app_id is required to associated it on one of those apps!"
                    ),
                )
            elif social_apps_count < 1:
                raise serializers.ValidationError(
                    _(
                        f"You don't have any Social App for this provider [{adapter.provider_id}]"
                    ),
                )
            else:
                # backward compatibility
                app = social_apps[0].legacy()

        request.app = app
        code = attrs.get("code")

        if code:
            self.set_callback_url(view=view, adapter_class=adapter_class)
            self.client_class = getattr(view, "client_class", None)

            if not self.client_class:
                raise serializers.ValidationError(
                    _("Define client_class in view"),
                )
            provider = adapter.get_provider()
            scope = provider.get_scope(request)
            client = self.client_class(
                request,
                app.client_id,
                app.secret,
                adapter.access_token_method,
                adapter.access_token_url,
                self.callback_url,
                scope,
                scope_delimiter=adapter.scope_delimiter,
                headers=adapter.headers,
                basic_auth=adapter.basic_auth,
            )
            attrs = client.get_access_token(code)

        if not attrs.get("access_token") and attrs.get("id_token"):
            attrs["access_token"] = attrs["id_token"]

        social_token = adapter.parse_token(attrs)
        social_token.app = app

        try:
            login = self.get_social_login(adapter, app, social_token, attrs)
            complete_social_login(request, login)
        except HTTPError:
            raise serializers.ValidationError(_("Incorrect value"))

        attrs["type"] = LoginType.EXISTING

        if not login.is_existing:
            login.lookup()
            try:
                user = get_user_model().objects.get(
                    email=login.user.email,
                )
                login.user = user
                attrs["type"] = LoginType.BINDING
            except ObjectDoesNotExist:
                pass
            login.save(request, connect=True)

        user = login.account.user
        attrs["user"] = user

        if not user.is_activated:
            user.is_activated = True
            if user.has_usable_password():
                attrs["type"] = LoginType.ACTIVATION
            else:
                if "name" in login.account.extra_data.keys():
                    user.name = login.account.extra_data.get("name", "")
                user.master = JsOrc.master(h=user._h, name=user.email).id
                attrs["type"] = LoginType.REGISTRATION

            user._h.commit()
            user.save()

        return attrs


class JSXSocialLoginView(SocialLoginView):
    adapter_class = None
    provider = None
    client_class = OAuth2Client
    serializer_class = JSXSocialLoginSerializer
    # permission_classes = [IsValidateLicense, AllowAny]

    def get_callback_url(self, request):
        callback_url = request.data.get("callback_url")
        if callback_url:
            return callback_url

        if self.provider:
            prov = PROVIDERS_MAPPING[self.provider]
            return lookup_global_config(
                name=prov["URL_KEY"],
                default=f'{request.build_absolute_uri("/")[:-1]}{prov["DEFAULT_REDIRECT_URI"]}',
            )
        raise RuntimeError(
            "Provider name cannot be empty or None. "
            'Please provide a valid provider name e.g. "GOOGLE"'
        )

    def post(self, request, *args, **kwargs):
        self.request = request
        self.request_log()

        self.callback_url = self.get_callback_url(request)
        self.serializer = self.get_serializer(data=self.request.data)
        self.serializer.is_valid(raise_exception=True)
        data = self.serializer.validated_data
        self.user = data.get("user")

        expiry = (
            knox_settings.TOKEN_TTL
            if knox_settings.TOKEN_TTL
            else timedelta(hours=settings.KNOX_TOKEN_EXPIRY)
        )
        expires_in: str = self.request.query_params.get("expires_in")

        if expires_in and expires_in.isnumeric():
            expiry = None if expires_in == "0" else timedelta(hours=int(expires_in))

        instance, token = AuthToken.objects.create(self.user, expiry)

        auth_user = authenticate(
            request=self.request, username=self.user.email, password=self.user.password
        )
        login_type = data.get("type")
        resp = {
            "email": auth_user.email if auth_user else self.user.email,
            "token": token,
            "exp": instance.expiry,
            "type": login_type.name,
            "details": login_type.value,
        }

        self.response_log(resp)
        return Response(resp)

    def request_log(self):
        request = self.request
        request.start_time = time()
        pl_peek = str(dict(request.data))[:256]
        user_agent = request.META.get("HTTP_USER_AGENT", "")
        extra = {"query": request.GET, "data": request.POST}
        extra["extra_fields"] = list(extra.keys())

        logger.info(
            f"Incoming call to {Cc.TG}{type(self).__name__}{Cc.EC} with {pl_peek} via {user_agent}",
            extra=extra,
        )

    def response_log(self, resp):
        request = self.request
        master = self.user.get_master()
        hook = master._h

        tot_time = time() - request.start_time
        save_count = len(hook.save_obj_list)
        touch_count = len(hook.mem.keys())
        db_touches = hook.db_touch_count
        red_touches = hook.red_touch_count
        touch_kb = hook.mem_size()

        res_peek = str(resp)[:256]
        log_str = str(
            f"API call from {Cc.TG}{master.name}{Cc.EC}:{master.jid}"
            f" to {Cc.TG}{type(self).__name__}{Cc.EC}"
            f" completed in {Cc.TY}{tot_time:.3f} seconds{Cc.EC}"
            f" touched {Cc.TY}{touch_count}{Cc.EC} mem /"
            f" {Cc.TY}{red_touches}{Cc.EC} redis /"
            f" {Cc.TY}{db_touches}{Cc.EC} db "
            f" ({Cc.TY}{touch_kb:.1f}kb{Cc.EC}) and"
            f" saving {Cc.TY}{save_count}{Cc.EC} objects."
            f" Response: {res_peek}."
        )

        log_dict = {
            "api_name": type(self).__name__,
            "request_latency": tot_time,
            "objects_touched": touch_count,
            "redis_touches": red_touches,
            "db_touches": db_touches,
            "objects_touched_size": touch_kb,
            "objects_saved": save_count,
            "caller_name": master.name,
            "caller_jid": master.jid,
        }
        try:
            api_result_str = dumps(resp)[:OBJECT_LOG_LIMIT]
        except TypeError:
            api_result_str = str(resp)[:OBJECT_LOG_LIMIT]
        log_dict["api_response"] = api_result_str
        log_dict["extra_fields"] = list(log_dict.keys())

        logger.info(log_str, extra=log_dict)



import os
import time
from django.contrib.auth import get_user_model
from django.urls import reverse
import json
from rest_framework.test import APIClient

from time import sleep
from jaseci.utils.utils import logger

APP_PATH = os.path.join(os.path.dirname(__file__), "example_jac")


class JsorcLoadTest:
    """
    A load tester module around JSORC.
    TODO: add experiment test cases
    """

    def __init__(self, test):
        self.client = APIClient()
        user_email = "JSCITfdfdEST_test@jaseci.com"
        suser_email = "JSCITfdfdEST_test2@jaseci.com"
        password = "password"
        try:
            self.user = get_user_model().objects.get(email=user_email.lower())
        except get_user_model().DoesNotExist:
            self.user = get_user_model().objects.create_user(user_email, password)
        try:
            self.suser = get_user_model().objects.get(email=suser_email.lower())
        except get_user_model().DoesNotExist:
            self.suser = get_user_model().objects.create_superuser(
                suser_email, password
            )

        self.auth_client = APIClient()
        self.auth_client.force_authenticate(self.user)
        self.sauth_client = APIClient()
        self.sauth_client.force_authenticate(self.suser)

        self.test = test

    def run_test(
        self, experiment, mem, policy, experiment_duration, eval_phase, perf_phase
    ):
        """
        Run the corresponding jsorc test
        """
        test_func = getattr(self, self.test)
        if experiment == "":
            return test_func()
        else:
            return test_func(
                experiment, mem, policy, experiment_duration, eval_phase, perf_phase
            )

    def load_action(self, name, mode, wait_for_ready=False):
        """
        Load action in corresponding mode.
        If wait is True, this function will block until the action is fully loaded,
        including microserivice is up and connected etc.
        """
        payload = {"op": "jsorc_actions_load", "name": name, "mode": mode}
        res = self.sauth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        while wait_for_ready:
            mode = "module" if mode == "local" else mode
            if res.data["action_status"]["mode"] == mode:
                break
            sleep(10)
            payload = {"op": "jsorc_actions_load", "name": name, "mode": mode}
            res = self.sauth_client.post(
                reverse(f'jac_api:{payload["op"]}'), payload, format="json"
            )

    def unload_action(self, name, mode, retire_svc=False):
        payload = {
            "op": "jsorc_actions_unload",
            "name": name,
            "mode": mode,
            "retire_svc": retire_svc,
        }
        res = self.sauth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        return res.data

    def run_walker(self, walker_name, ctx={}):
        payload = {"op": "walker_run", "name": walker_name, "ctx": ctx}
        res = self.sauth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        return res.data

    def sentinel_register(self, jac_file):
        jac_code = open(jac_file).read()
        # TODO: remove this once opt_level bug is fixed
        payload = {"op": "sentinel_register", "code": jac_code, "opt_level": 2}
        res = self.sauth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )

    def set_jsorc_actionpolicy(self, policy_name, policy_params):
        payload = {
            "op": "jsorc_actionpolicy_set",
            "policy_name": policy_name,
            "policy_params": policy_params,
        }
        res = self.sauth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )

    def start_benchmark(self):
        # Start benchmark
        payload = {"op": "jsorc_benchmark_start"}
        self.sauth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )

    def stop_benchmark(self):
        # Stop benchmark and get report
        payload = {"op": "jsorc_benchmark_stop", "report": True}
        res = self.sauth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        return res.data

    def start_actions_tracking(self):
        payload = {"op": "jsorc_trackact_start"}
        res = self.sauth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        return res.data

    def stop_actions_tracking(self):
        payload = {"op": "jsorc_trackact_stop"}
        res = self.sauth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        return res.data

    def load_action_config(self, action_pkg, action_module):
        """ """
        payload = {
            "op": "jsorc_actions_config",
            "config": action_pkg,
            "name": action_module,
        }
        res = self.sauth_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )
        return res.data

    def synthetic_apps(
        self, experiment, mem, policy, experiment_duration, eval_phase, perf_phase
    ):
        """
        Run synthetic application
        Available applications are in jaseci_serv/base/example_jac
        """
        try:
            results = {}
            node_mem = [int(mem) * 1024]
            apps = [experiment]
            policies = [policy]
            app_to_actions = {
                "zeroshot_faq_bot": ["jac_nlp.text_seg", "jac_nlp.use_qa"],
                "sentence_pairing": ["jac_nlp.sbert_sim", "jac_nlp.bi_enc"],
                "discussion_analysis": ["jac_nlp.bi_enc", "jac_nlp.cl_summer"],
                "flight_chatbot": ["jac_nlp.use_qa", "jac_nlp.ent_ext"],
                "restaurant_chatbot": ["jac_nlp.bi_enc", "jac_nlp.tfm_ner"],
                "virtual_assistant": [
                    "jac_nlp.text_seg",
                    "jac_nlp.bi_enc",
                    "jac_nlp.tfm_ner",
                    "jac_nlp.sbert_sim",
                    "jac_nlp.use_qa",
                ],
                "flow_analysis": [
                    "jac_nlp.text_seg",
                    "jac_nlp.tfm_ner",
                    "jac_nlp.use_enc",
                ],
                "weather_and_time_assitance": [
                    "jac_speech.vc_tts",
                    "jac_speech.stt",
                    "jac_nlp.bi_enc",
                ],
            }

            for app in apps:
                jac_file = os.path.join(APP_PATH, f"{app}.jac")
                self.sentinel_register(jac_file)
                action_modules = app_to_actions[app]
                for policy in policies:
                    if policy == "all_local" or policy == "all_remote":
                        policy_params = [{}]
                    else:
                        policy_params = [{"node_mem": nm} for nm in node_mem]
                    for pparams in policy_params:
                        pparams["eval_phase"] = eval_phase
                        pparams["perf_phase"] = perf_phase
                        if policy == "all_local":
                            jsorc_policy = "Default"
                            for module in action_modules:
                                package, module = module.split(".")
                                self.load_action_config(f"{package}.config", module)
                                self.load_action(module, "local", wait_for_ready=True)
                        elif policy == "all_remote":
                            jsorc_policy = "Default"
                            for module in action_modules:
                                package, module = module.split(".")
                                self.load_action_config(f"{package}.config", module)
                                self.load_action(module, "remote", wait_for_ready=True)
                        elif policy == "evaluation":
                            jsorc_policy = "Evaluation"
                            # For JSORC mode, we start as remote everything
                            for module in action_modules:
                                package, module = module.split(".")
                                self.load_action_config(f"{package}.config", module)
                                self.load_action(module, "remote", wait_for_ready=True)
                        else:
                            logger.error(f"Unrecognized policy {policy}")
                            return

                        self.set_jsorc_actionpolicy(jsorc_policy, policy_params=pparams)
                        #
                        # Experiment Start
                        #
                        self.start_benchmark()
                        self.start_actions_tracking()
                        start_ts = time.time()
                        while (time.time() - start_ts) < experiment_duration:
                            res = self.run_walker(app)
                        result = self.stop_benchmark()
                        action_result = self.stop_actions_tracking()
                        if policy == "all_local" or policy == "all_remote":
                            policy_str = policy
                        else:
                            policy_str = f"{policy}-mem-{pparams['node_mem']}"
                        results.setdefault(app, {})[policy_str] = {
                            "walker_level": result,
                            "action_level": action_result,
                        }
                        #
                        # Experiment Ends. Unload actions. Reset the cluster
                        #
                        if policy == "all_local":
                            for module in action_modules:
                                package, module = module.split(".")
                                self.unload_action(
                                    module, mode="local", retire_svc=True
                                )
                        elif policy == "all_remote":
                            for module in action_modules:
                                package, module = module.split(".")
                                self.unload_action(
                                    module, mode="remote", retire_svc=True
                                )
                        else:
                            for module in action_modules:
                                package, module = module.split(".")
                                self.unload_action(module, mode="auto", retire_svc=True)
                        sleep(10)
            self.set_jsorc_actionpolicy("Default", policy_params={})
            path = "/root/.jaseci/models/exp_results/"
            os.makedirs(path, exist_ok=True)
            with open(f"{path}/{app}_{policy}.json", "w") as fp:
                json.dump(results, fp)
            return results
        except Exception as e:
            return f"Exception: {e}"



from django.http import HttpResponseNotFound
from django.shortcuts import render
from django.template import TemplateDoesNotExist


def studio_index(request):
    try:
        return render(request, "studio/index.html")
    except TemplateDoesNotExist:
        return HttpResponseNotFound("Jaseci Studio is not installed.")


def studio_dashboard(request):
    try:
        return render(request, "studio/dashboard.html")
    except TemplateDoesNotExist:
        return HttpResponseNotFound("Jaseci Studio is not installed.")


def studio_graph_viewer(request):
    try:
        return render(request, "studio/graph-viewer.html")
    except TemplateDoesNotExist:
        return HttpResponseNotFound("Jaseci Studio is not installed.")


def studio_logs_viewer(request):
    try:
        return render(request, "studio/logs.html")
    except TemplateDoesNotExist:
        return HttpResponseNotFound("Jaseci Studio is not installed.")


def studio_architype(request):
    try:
        return render(request, "studio/architype.html")
    except TemplateDoesNotExist:
        return HttpResponseNotFound("Jaseci Studio is not installed.")


def studio_actions(request):
    try:
        return render(request, "studio/actions.html")
    except TemplateDoesNotExist:
        return HttpResponseNotFound("Jaseci Studio is not installed.")



import jwt
from time import time
from datetime import datetime
from rest_framework.test import APIClient
from jaseci.utils.utils import TestCaseHelper
from django.test import TestCase
from knox.models import AuthToken
from knox.settings import CONSTANTS
from allauth.socialaccount.providers.oauth2.client import OAuth2Client
from jaseci_serv.base.models import User
from jaseci_serv.jsx_oauth.models import SocialApp

mocked_access_token = "ya29.a0AX9GBdVvL b7JjkGJmFOy_oBhQC4jUdOKOjvpHGrcyFjCvWeqwW_1yTkWHJxA6jqEfXYkSzvwBCbq6XsA0o-fTDybsR9T8v8ilXM43IvqIQavWqxebTdjQo5ikPXDabrpJWQ3Uf1gMv1SfnwfeH6KdXxkGwPicaCgYKARwSARMSFQHUCsbCQepIVdu7SMuLZ5c-iI8Awg0163"
mocked_refresh_token = "1//0evz2w3jKj1qUCgYIARAAGA4SNwF-L9IrCiih-u1fKn9ECfKNMtKTK-y6wNZWcTfN8FO1oLnws4QqHkZZn2GSC-GUirdnOORUDIc"
mocked_code = (
    "4%2F0AWgavdcDFN1oZoRrqBlirNICcBprusdJpC0LPYnARjbWOgHGsxJAiw0hVD9zQL3y-rPwnA"
)
mocked_id_token = "eyJhbGciOiJSUzI1NiIsImtpZCI6IjAyYTYxZWZkMmE0NGZjMjE1MTQ4ZDRlZmZjMzRkNmE3YjJhYzI2ZjAiLCJ0eXAiOiJKV1QifQ.eyJpc3MiOiJhY2NvdW50cy5nb29nbGUuY29tIiwiYXpwIjoiMzg0NzUyNjExMTkwLWUwNGpiNDNpOGNuZnA5NTNyMW0xZjVtcTVhYmQ5MTlpLmFwcHMuZ29vZ2xldXNlcmNvbnRlbnQuY29tIiwiYXVkIjoiMzg0NzUyNjExMTkwLWUwNGpiNDNpOGNuZnA5NTNyMW0xZjVtcTVhYmQ5MTlpLmFwcHMuZ29vZ2xldXNlcmNvbnRlbnQuY29tIiwic3ViIjoiMTE2NjIwMzI4OTM4NjYyMjkwNDI3IiwiZW1haWwiOiJqYXNlY2kuZGV2QGdtYWlsLmNvbSIsImVtYWlsX3ZlcmlmaWVkIjp0cnVlLCJhdF9oYXNoIjoiWUd5d2tPZG1VVEo5dU1ma3ZxRGNPZyIsImlhdCI6MTY3MTcwMTQ0MiwiZXhwIjoxNjcxNzA1MDQyfQ.jR23tvfhJYjxHVzOZsdNphZ_2vJ4iPIraY_cbwM9NerfmdiBEhGxhhmksF8X2K5G4Yg607StTjhKNE0XStiWhhBzZu3w7ArujqPoL2NgT2AR8kyQ91cGZcl6REKwhZBQHc7oX9V-UmK6wo8bWfPgdZ0xq_D_YDFaJuabph9d7YutGy4dsEjMkgWN-E1IdVXlXPBciaBK6Djx06x4Wr_dlfeZl0KWHcxWTDXS1DV3rEFpyZlqSTYsTVdnP-nRJZqXbgN7IRNgJIehxBleNRoXCvmm5rD5D1NCI84OOplUEYlpH_dNF8iSdrh7GLgijKl6kbKGlbWyrrBYeJa07nnvJw"


def mocked_get_access_token(*ignored):
    return {
        "access_token": mocked_access_token,
        "expires_in": 3599,
        "refresh_token": mocked_refresh_token,
        "scope": "openid https://www.googleapis.com/auth/userinfo.email https://www.googleapis.com/auth/userinfo.profile",
        "token_type": "Bearer",
        "id_token": mocked_id_token,
    }


def mocked_decode(*ignored_args, **ignored_kwargs):
    t = int(time())
    return {
        "iss": "https://accounts.google.com",
        "azp": "384752611190-e04jb43i8cnfp953r1m1f5mq5abd919i.apps.googleusercontent.com",
        "aud": "384752611190-e04jb43i8cnfp953r1m1f5mq5abd919i.apps.googleusercontent.com",
        "sub": "116620328938662290427",
        "email": "jaseci.dev@gmail.com",
        "email_verified": True,
        "at_hash": "jCcj5Kh-b6LcfHEIVKFJDg",
        "name": "jaseci dev",
        "picture": "https://lh3.googleusercontent.com/a/AEdFTp5FRawOuoF9Tl5jYkExTn1kGbDXqNoJP9ejF6No=s96-c",
        "given_name": "jaseci",
        "family_name": "dev",
        "locale": "en-US",
        "iat": t,
        "exp": t + 1000000,
    }


class SocialAuthTest(TestCaseHelper, TestCase):
    """Test the Social Authentication"""

    def setUp(self):
        super().setUp()
        self.client = APIClient()
        OAuth2Client.get_access_token = mocked_get_access_token
        jwt.decode = mocked_decode
        self.google_app = SocialApp.objects.create(
            provider="google",
            name="Google OAuth",
            client_id="384752611190-e04jb43i8cnfp953r1m1f5mq5abd919i.apps.googleusercontent.com",
            secret="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",
        )
        self.google_app.save()

    def tearDown(self):
        super().tearDown()

    def create_google_provider(self):
        google_app = SocialApp.objects.create(
            provider="google",
            name="Google OAuth 2",
            client_id="384752611190-e04jb43i8cnfp953r1m1f5mq5abd919i.apps.googleusercontent.com",
            secret="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",
        )

        google_app.save()

        return google_app

    def request_social_sign_in_with_expires_in(self, expires_in_query=""):
        now = datetime.now()

        token = self.client.post(
            path=f"/auth/google/?{expires_in_query}",
            data={"code": mocked_code},
        ).data["token"]

        auth_token = AuthToken.objects.get(
            token_key=token[: CONSTANTS.TOKEN_KEY_LENGTH]
        )

        # convert to hour count if available
        return (
            int((auth_token.expiry - now).total_seconds() / 60 / 60)
            if auth_token.expiry
            else None
        )

    def request_social_sign_in_with_multiple_provider(self, data):
        res = self.client.post(
            path="/auth/google/",
            data=data,
        ).json()

        self.assertEqual(
            {
                "non_field_errors": [
                    "You have multiple google Social App. app_id is required to associated it on one of those apps!"
                ]
            },
            res,
        )
        self.assertEqual(0, User.objects.count())

    def request_social_sign_in(self, data):
        # no user yet
        self.assertEqual(0, User.objects.count())

        res = self.client.post(
            path="/auth/google/",
            data=data,
        ).data

        # user should be created
        self.assertEqual(1, User.objects.count())
        created_user = User.objects.first()

        self.client.credentials(HTTP_AUTHORIZATION="token " + res["token"])

        # token should work
        res = self.client.get(path="/user/manage/").data

        self.assertEqual(
            res,
            {
                "id": 1,
                "email": "jaseci.dev@gmail.com",
                "name": "jaseci dev",
                "is_activated": True,
                "is_superuser": False,
            },
        )

        self.assertEqual(created_user.id, res["id"])
        self.assertEqual(created_user.email, res["email"])
        self.assertEqual(created_user.name, res["name"])
        self.assertEqual(created_user.is_activated, res["is_activated"])
        self.assertIsNotNone(created_user.get_master())

        # getting token again should not create another user
        res = self.client.post(
            path="/auth/google/",
            data=data,
        ).data

        # user count should still be the same
        self.assertEqual(1, User.objects.count())
        self.assertEqual(created_user, User.objects.first())

        User.objects.all().delete()

    def test_social_auth_flow_using_authorization_code(self):
        """Test social auth flow from 'with autorization code'"""
        self.request_social_sign_in({"code": mocked_code})

    def test_social_auth_flow_using_id_token(self):
        """Test social auth flow from 'with id token'"""
        self.request_social_sign_in({"id_token": mocked_id_token})

    def test_social_auth_flow_using_access_token(self):
        """Test social auth flow from 'with access token'"""
        self.request_social_sign_in(mocked_get_access_token())

    def test_social_auth_flow_with_multiple_provider(self):
        """Test social auth flow from 'with autorization code'"""

        self.create_google_provider()

        self.request_social_sign_in_with_multiple_provider(
            {
                "code": mocked_code,
            }
        )

        self.request_social_sign_in_with_multiple_provider(
            {"id_token": mocked_id_token}
        )

        self.request_social_sign_in_with_multiple_provider(mocked_get_access_token())

    def test_social_auth_flow_using_authorization_code_with_app_id(self):
        self.request_social_sign_in(
            {"app_id": self.create_google_provider().id, "code": mocked_code}
        )

    def test_social_auth_flow_using_id_token_with_app_id(self):
        self.request_social_sign_in(
            {
                "app_id": self.create_google_provider().id,
                "id_token": mocked_id_token,
            }
        )

    def test_social_auth_flow_using_access_token_with_app_id(self):
        data = mocked_get_access_token()
        data["app_id"] = self.create_google_provider().id
        self.request_social_sign_in(data)

    def test_social_auth_flow_with_expires_in_param(self):
        # no expires_in_param: defaults to 12 hrs
        self.assertEqual(12, self.request_social_sign_in_with_expires_in())

        # expires_in: 24 hrs
        self.assertEqual(
            24, self.request_social_sign_in_with_expires_in("expires_in=24")
        )

        # no expiry
        self.assertEqual(
            None, self.request_social_sign_in_with_expires_in("expires_in=0")
        )



from uuid import uuid4
from django.db import models
from django.utils.translation import gettext_lazy as _
from allauth.socialaccount.models import SocialApp as LegacySocialApp
from allauth.socialaccount import providers


class SocialLoginProvider(models.TextChoices):
    GOOGLE = "GOOGLE", _("Google")
    FACEBOOK = "FACEBOOK", _("Facebook")
    MICROSOFT = "MICROSOFT", _("Microsoft")
    GITHUB = "GITHUB", _("GitHub")
    OKTA = "OKTA", _("Okta")
    OPENID = "OPENID", _("OpenID")
    APPLE = "APPLE", _("Apple")


PROVIDERS_MAPPING = {
    SocialLoginProvider.GOOGLE: {
        "URL_KEY": SocialLoginProvider.GOOGLE + "_REDIRECT_URI",
        "LOGIN_URL": "https://accounts.google.com/o/oauth2/v2/auth?redirect_uri={callback_url}&prompt=consent&response_type=code&client_id={client_id}&scope=https%3A%2F%2Fwww.googleapis.com%2Fauth%2Fuserinfo.email%20https%3A%2F%2Fwww.googleapis.com%2Fauth%2Fuserinfo.profile%20openid",
        "DEFAULT_REDIRECT_URI": "/auth/examples/google/",
    },
    SocialLoginProvider.FACEBOOK: {
        "URL_KEY": SocialLoginProvider.FACEBOOK + "_REDIRECT_URI",
        "LOGIN_URL": "https://www.facebook.com/dialog/oauth/?client_id={client_id}&redirect_uri={callback_url}&scope=email&state={state}",
        "DEFAULT_REDIRECT_URI": "/auth/examples/facebook/",
    },
    SocialLoginProvider.GITHUB: {
        "URL_KEY": SocialLoginProvider.GITHUB + "_REDIRECT_URI",
        "LOGIN_URL": "https://github.com/login/oauth/authorize?client_id={client_id}&redirect_uri={callback_url}&scope=read:user,user:email&state={state}",
        "DEFAULT_REDIRECT_URI": "/auth/examples/github/",
    },
    SocialLoginProvider.MICROSOFT: {
        "URL_KEY": SocialLoginProvider.MICROSOFT + "_REDIRECT_URI",
        "LOGIN_URL": "",
        "DEFAULT_REDIRECT_URI": "/auth/examples/microsoft/",
    },
    SocialLoginProvider.OKTA: {
        "URL_KEY": SocialLoginProvider.OKTA + "_REDIRECT_URI",
        "LOGIN_URL": "",
        "DEFAULT_REDIRECT_URI": "/auth/examples/okta/",
    },
    SocialLoginProvider.APPLE: {
        "URL_KEY": SocialLoginProvider.APPLE + "_REDIRECT_URI",
        "LOGIN_URL": "",
        "DEFAULT_REDIRECT_URI": "/auth/examples/apple/",
    },
}


class SocialApp(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid4, editable=False)
    name = models.CharField(verbose_name=_("name"), max_length=40)
    provider = models.CharField(
        verbose_name=_("provider"),
        max_length=30,
        choices=providers.registry.as_choices(),
    )
    client_id = models.CharField(
        verbose_name=_("client id"),
        max_length=191,
        help_text=_("App ID, or consumer key"),
    )
    secret = models.CharField(
        verbose_name=_("secret key"),
        max_length=191,
        blank=True,
        help_text=_("API secret, client secret, or consumer secret"),
    )
    key = models.CharField(
        verbose_name=_("key"), max_length=191, blank=True, help_text=_("Key")
    )

    certificate_key = models.TextField(verbose_name=_("certificate key"), blank=True)

    class Meta:
        verbose_name = _("social application")
        verbose_name_plural = _("social applications")

    def __str__(self):
        return self.name

    def legacy(self):
        app = LegacySocialApp(
            provider=self.provider,
            name=self.name,
            client_id=self.client_id,
            secret=self.secret,
            key=self.key,
        )
        app.certificate_key = self.certificate_key
        return app



OAUTH_APPS = [
    "allauth.socialaccount.providers.google",
    "allauth.socialaccount.providers.facebook",
    "allauth.socialaccount.providers.github",
    "allauth.socialaccount.providers.apple",
]

SOCIALACCOUNT_EMAIL_VERIFICATION = "none"

KNOX_TOKEN_EXPIRY = 12

REST_AUTH_TOKEN_MODEL = None

ACCOUNT_USER_MODEL_USERNAME_FIELD = None
ACCOUNT_EMAIL_REQUIRED = True
ACCOUNT_USERNAME_REQUIRED = False
ACCOUNT_AUTHENTICATION_METHOD = "email"



from django.contrib.auth import get_user_model
from django.urls import reverse

from rest_framework.test import APIClient

from jaseci.utils.utils import TestCaseHelper
from django.test import TestCase

import os


class PublicWalkerApiTests(TestCaseHelper, TestCase):
    """Test the authorized user node API"""

    def setUp(self):
        super().setUp()
        get_user_model().objects.create_user(
            "throwawayJSCITfdfdEST_test@jaseci.com", "password"
        )
        self.user = get_user_model().objects.create_user(
            "JSCITfdfdEST_test@jaseci.com", "password"
        )
        self.user2 = get_user_model().objects.create_user(
            "JSCITfdfdEST_test1@jaseci.com", "password"
        )
        self.admin = get_user_model().objects.create_superuser(
            "JSCITfdfdEST_test2@jaseci.com", "password"
        )
        self.no_user_client = APIClient()
        self.user_client = APIClient()
        self.user2_client = APIClient()
        self.admin_client = APIClient()
        self.user_client.force_authenticate(self.user)
        self.user2_client.force_authenticate(self.user2)
        self.admin_client.force_authenticate(self.admin)

        self.master = self.user.get_master()

        zsb_file = open(os.path.dirname(__file__) + "/public.jac").read()

        payload = {"op": "sentinel_register", "name": "public", "code": zsb_file}
        self.admin_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )

        payload = {"op": "global_sentinel_set"}
        self.admin_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )

        payload = {"op": "sentinel_active_global"}
        self.admin_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )

        payload = {"op": "sentinel_register", "name": "public", "code": zsb_file}
        self.user_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )

        payload = {"op": "sentinel_register", "name": "public", "code": zsb_file}
        self.user2_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        )

        payload = {"op": "walker_run", "name": "manual_init"}
        res = self.user_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        ).data
        self.user_node = res["report"][0]["jid"]

        payload = {"op": "walker_run", "name": "manual_init"}
        res = self.user2_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        ).data
        self.user2_node = res["report"][0]["jid"]

        payload = {"op": "walker_run", "name": "manual_init"}
        res = self.admin_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        ).data
        self.admin_node = res["report"][0]["jid"]

        payload = {"op": "walker_spawn_create", "name": "show_a_details"}
        res = self.user_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        ).data
        self.user_walker = res["jid"].split(":")[2]

        payload = {"op": "walker_spawn_create", "name": "show_a_details"}
        res = self.user2_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        ).data
        self.user2_walker = res["jid"].split(":")[2]

        payload = {"op": "walker_spawn_create", "name": "show_a_details"}
        res = self.admin_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        ).data
        self.admin_walker = res["jid"].split(":")[2]

        payload = {
            "op": "walker_get",
            "mode": "keys",
            "wlk": "spawned:walker:show_a_details",
        }
        res = self.user_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        ).data

        self.user_key = res["requestor"]

        payload = {
            "op": "walker_get",
            "mode": "keys",
            "wlk": "spawned:walker:show_a_details",
        }
        res = self.user2_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        ).data

        self.user2_key = res["requestor"]

        payload = {
            "op": "walker_get",
            "mode": "keys",
            "wlk": "spawned:walker:show_a_details",
        }
        res = self.admin_client.post(
            reverse(f'jac_api:{payload["op"]}'), payload, format="json"
        ).data

        self.admin_key = res["requestor"]

    def tearDown(self):
        super().tearDown()

    def walker_callback_trigger(self, payload):
        payload["op"] = "walker_callback"
        return self.no_user_client.post(
            reverse(f'jac_api:{payload["op"]}', args=[payload["nd"], payload["wlk"]])
            + "?keys="
            + payload["key"],
            payload,
            format="json",
        ).data

    def test_user_public_walker_to_user_node(self):
        """Test public API for walker callback"""

        res = self.walker_callback_trigger(
            {"key": self.user_key, "wlk": self.user_walker, "nd": self.user_node}
        )

        self.assertEqual(res["report"][0]["context"]["value"], "user")
        self.assertEqual(res["report"][1]["context"]["value"], "changed")

    def test_user_public_walker_to_admin_node(self):
        """Test public API for walker callback"""

        res = self.walker_callback_trigger(
            {"key": self.user_key, "wlk": self.user_walker, "nd": self.admin_node}
        )

        self.assertFalse(res["success"])

    def test_user_public_walker_to_other_user_node(self):
        """Test public API for walker callback"""

        res = self.walker_callback_trigger(
            {"key": self.user_key, "wlk": self.user_walker, "nd": self.user2_node}
        )

        self.assertFalse(res["success"])

        res = self.walker_callback_trigger(
            {"key": self.user2_key, "wlk": self.user2_walker, "nd": self.user_node}
        )

        self.assertFalse(res["success"])

    def test_admin_public_walker_to_user_node(self):
        """Test public API for walker callback"""

        res = self.walker_callback_trigger(
            {"key": self.admin_key, "wlk": self.admin_walker, "nd": self.user_node}
        )

        self.assertEqual(res["report"][0]["context"]["value"], "user")
        self.assertEqual(res["report"][1]["context"]["value"], "changed")

        res = self.walker_callback_trigger(
            {"key": self.admin_key, "wlk": self.admin_walker, "nd": self.user2_node}
        )

        self.assertEqual(res["report"][0]["context"]["value"], "user")
        self.assertEqual(res["report"][1]["context"]["value"], "changed")
